# -*- coding: utf-8 -*-

import argparse
import datetime
import os
import json
import logging
import sys

import boto3

from tabulate import tabulate

from s4 import sync
from s4 import utils
from s4.clients import local, s3


VERSION = '0.1.21'
CONFIG_FOLDER_PATH = os.path.expanduser('~/.config/s4')
CONFIG_FILE_PATH = os.path.join(CONFIG_FOLDER_PATH, 'sync.conf')


def get_s3_client(target, aws_access_key_id, aws_secret_access_key, region_name):
    s3_uri = s3.parse_s3_uri(target)
    s3_client = boto3.client(
        's3',
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        region_name=region_name,
    )
    return s3.S3SyncClient(s3_client, s3_uri.bucket, s3_uri.key)


def get_local_client(target):
    return local.LocalSyncClient(target)


def main(arguments):
    parser = argparse.ArgumentParser(prog='s4')
    parser.add_argument(
        '--log-level',
        default='INFO',
        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR'],
    )
    subparsers = parser.add_subparsers(dest='command')

    subparsers.add_parser('version')

    sync_parser = subparsers.add_parser('sync')
    sync_parser.add_argument('targets', nargs='*')
    sync_parser.add_argument('--conflicts', default=None, choices=['1', '2', 'ignore'])

    subparsers.add_parser('targets')
    subparsers.add_parser('add')
    edit_parser = subparsers.add_parser('edit')
    edit_parser.add_argument('target')

    ls_parser = subparsers.add_parser('ls')
    ls_parser.add_argument('target')
    ls_parser.add_argument('--sort-by', '-s', choices=['key', 'local', 's3'], default='key')
    ls_parser.add_argument('--descending', '-d', action='store_true')
    ls_parser.add_argument(
        '--all', '-A',
        dest='show_all',
        action='store_true',
        help='show deleted files',
    )

    remove_parser = subparsers.add_parser('rm')
    remove_parser.add_argument('target')

    args = parser.parse_args(arguments)

    if args.log_level == 'DEBUG':
        log_format = '%(levelname)s:%(module)s:%(lineno)s %(message)s'
    else:
        log_format = '%(message)s'

    logging.basicConfig(format=log_format, level=args.log_level)

    # shut boto up
    logging.getLogger('boto3').setLevel(logging.CRITICAL)
    logging.getLogger('botocore').setLevel(logging.CRITICAL)
    logging.getLogger('nose').setLevel(logging.CRITICAL)
    logging.getLogger('s3transfer').setLevel(logging.CRITICAL)

    logger = logging.getLogger(__name__)
    logger.setLevel(args.log_level)

    config = get_config()

    try:
        if args.command == 'version':
            print(VERSION)
        elif args.command == 'sync':
            sync_command(args, config, logger)
        elif args.command == 'targets':
            targets_command(args, config, logger)
        elif args.command == 'add':
            add_command(args, config, logger)
        elif args.command == 'edit':
            edit_command(args, config, logger)
        elif args.command == 'ls':
            ls_command(args, config, logger)
        elif args.command == 'rm':
            rm_command(args, config, logger)
        else:
            parser.print_help()
    except KeyboardInterrupt:
        pass


def get_config():
    if not os.path.exists(CONFIG_FILE_PATH):
        return {}

    with open(CONFIG_FILE_PATH, 'r') as fp:
        config = json.load(fp)
    return config


def set_config(config):
    if not os.path.exists(CONFIG_FOLDER_PATH):
        os.makedirs(CONFIG_FOLDER_PATH)

    with open(CONFIG_FILE_PATH, 'w') as fp:
        json.dump(config, fp, indent=4)


def get_clients(entry):
    target_1 = entry['local_folder']
    target_2 = entry['s3_uri']
    aws_access_key_id = entry['aws_access_key_id']
    aws_secret_access_key = entry['aws_secret_access_key']
    region_name = entry['region_name']

    # append trailing slashes to prevent incorrect prefix matching on s3
    if not target_1.endswith('/'):
        target_1 += '/'
    if not target_2.endswith('/'):
        target_2 += '/'

    client_1 = get_local_client(target_1)
    client_2 = get_s3_client(target_2, aws_access_key_id, aws_secret_access_key, region_name)
    return client_1, client_2


def sync_command(args, config, logger):
    all_targets = list(config['targets'].keys())
    if not args.targets:
        targets = all_targets
    else:
        targets = args.targets

    try:
        for name in sorted(targets):
            if name not in config['targets']:
                logger.info('"%s" is an unknown target. Choices are: %s', name, all_targets)
                continue

            entry = config['targets'][name]
            client_1, client_2 = get_clients(entry)

            worker = sync.SyncWorker(client_1, client_2)

            logger.info('Syncing %s [%s <=> %s]', name, client_1.get_uri(), client_2.get_uri())
            worker.sync(conflict_choice=args.conflicts)
    except KeyboardInterrupt:
        logger.warning('Quitting due to Keyboard Interrupt...')


def targets_command(args, config, logger):
    if 'targets' not in config:
        return

    for name in sorted(config['targets']):
        entry = config['targets'][name]
        logger.info('%s: [%s <=> %s]', name, entry['local_folder'], entry['s3_uri'])


def add_command(args, config, logger):
    entry = {}

    entry['local_folder'] = os.path.expanduser(utils.get_input('local folder: '))
    entry['s3_uri'] = utils.get_input('s3 uri: ')
    entry['aws_access_key_id'] = utils.get_input('AWS Access Key ID: ')
    entry['aws_secret_access_key'] = utils.get_input('AWS Secret Access Key: ', secret=True)
    entry['region_name'] = utils.get_input('region name: ')

    default_name = os.path.basename(entry['s3_uri'])
    name = utils.get_input('Provide a name for this entry [{}]: '.format(default_name))

    if not name:
        name = default_name

    if 'targets' not in config:
        config['targets'] = {}

    config['targets'][name] = entry

    set_config(config)


def edit_command(args, config, logger):
    if 'targets' not in config:
        logger.info('You have not added any targets yet')
        logger.info('Use the "add" command to do this')
        return

    if args.target not in config['targets']:
        all_targets = sorted(list(config['targets'].keys()))
        logger.info('"%s" is an unknown target', args.target)
        logger.info('Choices are: %s', all_targets)
        return

    entry = config['targets'][args.target]

    local_folder = entry.get('local_folder', '')
    s3_uri = entry.get('s3_uri', '')
    aws_access_key_id = entry.get('aws_access_key_id')
    aws_secret_access_key = entry.get('aws_secret_access_key')
    region_name = entry.get('region_name')

    new_local_folder = utils.get_input('local folder [{}]: '.format(local_folder))
    new_s3_uri = utils.get_input('s3 uri [{}]: '.format(s3_uri))
    new_aws_access_key_id = utils.get_input('AWS Access Key ID [{}]: '.format(aws_access_key_id))

    secret_key_prompt = 'AWS Secret Access Key [{}]: '.format(aws_secret_access_key)
    new_aws_secret_access_key = utils.get_input(secret_key_prompt, secret=True)

    new_region_name = utils.get_input('region name [{}]: '.format(region_name))

    if new_local_folder:
        entry['local_folder'] = new_local_folder
    if new_s3_uri:
        entry['s3_uri'] = new_s3_uri
    if new_aws_access_key_id:
        entry['aws_access_key_id'] = new_aws_access_key_id
    if new_aws_secret_access_key:
        entry['aws_secret_access_key'] = new_aws_secret_access_key
    if new_region_name:
        entry['region_name'] = new_region_name

    config['targets'][args.target] = entry

    set_config(config)


def ls_command(args, config, logger):
    if 'targets' not in config:
        logger.info('You have not added any targets yet')
        logger.info('Use the "add" command to do this')
        return
    if args.target not in config['targets']:
        all_targets = sorted(list(config['targets'].keys()))
        logger.info('"%s" is an unknown target', args.target)
        logger.info('Choices are: %s', all_targets)
        return

    target = config['targets'][args.target]
    client_1, client_2 = get_clients(target)

    sort_by = args.sort_by.lower()
    descending = args.descending

    keys = set(client_1.index) | set(client_2.index)

    data = []
    for key in sorted(keys):
        entry_1 = client_1.index.get(key, {})
        entry_2 = client_2.index.get(key, {})

        ts_1 = entry_1.get('local_timestamp')
        ts_2 = entry_2.get('local_timestamp')

        if args.show_all or ts_1 is not None:
            data.append((
                key,
                datetime.datetime.utcfromtimestamp(int(ts_1)) if ts_1 is not None else '<deleted>',
                datetime.datetime.utcfromtimestamp(int(ts_2)) if ts_2 is not None else None,
            ))

    headers = ['key', 'local', 's3']
    data = sorted(data, reverse=descending, key=lambda x: x[headers.index(sort_by)])

    logger.info(tabulate(data, headers=headers))


def rm_command(args, config, logger):
    if 'targets' not in config:
        logger.info('You have not added any targets yet')
        return
    if args.target not in config['targets']:
        all_targets = sorted(list(config['targets'].keys()))
        logger.info('"%s" is an unknown target', args.target)
        logger.info('Choices are: %s', all_targets)
        return

    del config['targets'][args.target]
    set_config(config)


if __name__ == '__main__':
    main(sys.argv[1:])



# -*- coding: utf-8 -*-

import datetime


class SyncState(object):
    UPDATED = 'UPDATED'
    CREATED = 'CREATED'
    DELETED = 'DELETED'
    CONFLICT = 'CONFLICT'
    NOCHANGES = 'NOCHANGES'
    DOESNOTEXIST = 'DOESNOTEXIST'

    def __init__(self, state, local_timestamp, remote_timestamp):
        self.state = state
        self.local_timestamp = local_timestamp
        self.remote_timestamp = remote_timestamp

    def get_local_datetime(self):
        if self.local_timestamp is not None:
            return datetime.datetime.utcfromtimestamp(self.local_timestamp)
        else:
            return None

    def get_remote_datetime(self):
        if self.remote_timestamp is not None:
            return datetime.datetime.utcfromtimestamp(self.remote_timestamp)
        else:
            return None

    def __eq__(self, other):
        if not isinstance(other, SyncState):
            return False
        return (
            self.state == other.state and
            self.local_timestamp == other.local_timestamp and
            self.remote_timestamp == other.remote_timestamp
        )

    def __repr__(self):
        return 'SyncState<{}, local={}, remote={}>'.format(
            self.state, self.get_local_datetime(), self.get_remote_datetime()
        )


class SyncObject(object):
    def __init__(self, fp, total_size, timestamp):
        self.fp = fp
        self.total_size = total_size
        self.timestamp = timestamp

    def __repr__(self):
        return 'SyncObject<{}, {}, {}>'.format(self.fp, self.total_size, self.timestamp)


def get_sync_state(index_local, real_local, remote):
    # convert to int because not all clients support float precision
    index_local = int(index_local) if index_local is not None else None
    real_local = int(real_local) if real_local is not None else None
    remote = int(remote) if remote is not None else None

    if index_local is None and real_local:
        return SyncState(SyncState.CREATED, real_local, remote)
    elif real_local is None and index_local:
        return SyncState(SyncState.DELETED, real_local, remote)
    elif real_local is None and index_local is None and remote:
        return SyncState(SyncState.DELETED, real_local, remote)
    elif real_local is None and index_local is None:
        # Does not exist in this case, so no action to perform
        return SyncState(SyncState.DOESNOTEXIST, None, None)

    elif index_local < real_local:
        return SyncState(SyncState.UPDATED, real_local, remote)
    elif index_local > real_local:
        return SyncState(SyncState.CONFLICT, index_local, remote)   # corruption?
    else:
        return SyncState(SyncState.NOCHANGES, real_local, remote)


class SyncClient(object):
    def get_client_name(self):
        raise NotImplementedError()

    def get_uri(self, key=''):
        raise NotImplementedError()

    def put(self, key, sync_object):
        raise NotImplementedError()

    def get(self, key):
        raise NotImplementedError()

    def delete(self, key):
        raise NotImplementedError()

    def get_local_keys(self):
        raise NotImplementedError()

    def get_real_local_timestamp(self, key):
        raise NotImplementedError()

    def get_index_keys(self):
        raise NotImplementedError()

    def get_index_local_timestamp(self, key):
        raise NotImplementedError()

    def set_index_local_timestamp(self, key, timestamp):
        raise NotImplementedError()

    def get_remote_timestamp(self, key):
        raise NotImplementedError()

    def set_remote_timestamp(self, key, timestamp):
        raise NotImplementedError()

    def get_all_remote_timestamps(self):
        raise NotImplementedError()

    def get_all_index_local_timestamps(self):
        raise NotImplementedError()

    def get_all_real_local_timestamps(self):
        raise NotImplementedError()

    def get_all_keys(self):
        local_keys = self.get_local_keys()
        index_keys = self.get_index_keys()
        return list(set(local_keys) | set(index_keys))

    def update_index(self):
        keys = self.get_all_keys()
        index = {}

        for key in keys:
            index[key] = {
                'remote_timestamp': self.get_remote_timestamp(key),
                'local_timestamp': self.get_real_local_timestamp(key),
            }
        self.index = index

    def update_index_entry(self, key):
        self.index[key] = {
            'remote_timestamp': self.get_remote_timestamp(key),
            'local_timestamp': self.get_real_local_timestamp(key),
        }

    def flush_index(self):
        raise NotImplementedError()

    def get_action(self, key):
        """
        returns the action to perform on this key based on its
        state before the last sync.
        """
        index_local_timestamp = self.get_index_local_timestamp(key)
        real_local_timestamp = self.get_real_local_timestamp(key)
        remote_timestamp = self.get_remote_timestamp(key)
        return get_sync_state(index_local_timestamp, real_local_timestamp, remote_timestamp)

    def get_all_actions(self):
        real_local_timestamps = self.get_all_real_local_timestamps()
        index_local_timestamps = self.get_all_index_local_timestamps()
        remote_timestamps = self.get_all_remote_timestamps()

        keys = set(real_local_timestamps) | set(index_local_timestamps)

        results = {}
        for key in keys:
            index_local_timestamp = index_local_timestamps.get(key)
            real_local_timestamp = real_local_timestamps.get(key)
            remote_timestamp = remote_timestamps.get(key)

            results[key] = get_sync_state(
                index_local_timestamp,
                real_local_timestamp,
                remote_timestamp,
            )

        return results



# -*- coding: utf-8 -*-
import datetime
import json
import os

import freezegun


class InterruptedBytesIO(object):
    """
    Test helper class that imitates a BytesIO stream. Will return a stream of 0s for
    every requested read. When the number of times read was read reaches the parameter
    value of `interrupt_at`, a ValueError will be raised to simulate an interrupted IO
    transfer that could have occured
    """
    def __init__(self, interrupt_at=1):
        self.index = 0
        self.interrupt_at = interrupt_at

    def read(self, size):
        if self.index == self.interrupt_at:
            raise ValueError('Interrupted IO')
        else:
            self.index += 1
            return b'0' * size


def write_local(path, data=''):
    parent = os.path.dirname(path)
    if not os.path.exists(parent):
        os.makedirs(parent)
    with open(path, 'w') as fp:
        fp.write(data)


def get_local_contents(client, key):
    with open(os.path.join(client.path, key), 'rb') as fp:
        return fp.read()


def set_local_contents(client, key, timestamp=None, data=''):
    path = os.path.join(client.path, key)
    write_local(path, data)
    if timestamp is not None:
        os.utime(path, (timestamp, timestamp))


def delete_local(client, key):
    os.remove(os.path.join(client.path, key))


def set_local_index(client, data):
    with open(client.index_path(), 'w') as fp:
        json.dump(data, fp)
    client.reload_index()


def write_s3(boto, bucket, key, data=''):
    boto.put_object(
        Bucket=bucket,
        Key=key,
        Body=data,
    )


def set_s3_contents(s3_client, key, timestamp=None, data=''):
    if timestamp is None:
        freeze_time = datetime.datetime.utcnow()
    else:
        freeze_time = datetime.datetime.utcfromtimestamp(timestamp)

    with freezegun.freeze_time(freeze_time):
        write_s3(s3_client.boto, s3_client.bucket, os.path.join(s3_client.prefix, key), data)


def set_s3_index(s3_client, data):
    s3_client.boto.put_object(
        Bucket=s3_client.bucket,
        Key=os.path.join(s3_client.prefix, '.index'),
        Body=json.dumps(data),
    )
    s3_client.reload_index()



# -*- coding: utf-8 -*-
import datetime
import io
import os

import boto3

from botocore.exceptions import ClientError

import freezegun

from moto import mock_s3

import pytest

from s4.clients import s3, SyncObject
from s4.utils import to_timestamp
import utils


class TestParseS3URI(object):
    def test_empty(self):
        assert s3.parse_s3_uri('') is None

    def test_incorrect_format(self):
        assert s3.parse_s3_uri('nos3infront/some/path') is None
        assert s3.parse_s3_uri(':/232red+-32') is None

    def test_no_path(self):
        assert s3.parse_s3_uri('s3://bucket') is None

    def test_correct_output(self):
        actual_output = s3.parse_s3_uri('s3://fruit.bowl/apples/and/oranges/')
        assert actual_output.bucket == 'fruit.bowl'
        assert actual_output.key == 'apples/and/oranges/'


class TestIsIgnoredFile(object):
    def test_empty(self):
        assert s3.is_ignored_key('foo/bar/baz', []) is False

    def test_subdirectory(self):
        assert s3.is_ignored_key('foo/.git/baz', ['.git']) is True

    def test_file(self):
        assert s3.is_ignored_key('foo/ignoreme', ['ignore*']) is True


class TestS3SyncClient(object):
    def test_get_client_name(self, s3_client):
        assert s3_client.get_client_name() == 's3'

    def test_repr(self):
        boto_client = boto3.client(
            's3',
            aws_access_key_id='',
            aws_secret_access_key='',
            aws_session_token='',
        )

        client = s3.S3SyncClient(boto_client, 'testbucket', 'foo/bar')
        assert repr(client) == 'S3SyncClient<testbucket, foo/bar>'

    @mock_s3
    def test_get_uri(self):
        s3_client = boto3.client(
            's3',
            aws_access_key_id='',
            aws_secret_access_key='',
            aws_session_token='',
        )

        client = s3.S3SyncClient(s3_client, 'testbucket', 'foo/bar')
        assert client.get_uri() == 's3://testbucket/foo/bar/'
        assert client.get_uri('apples.txt') == 's3://testbucket/foo/bar/apples.txt'

    @mock_s3
    def test_index_path(self):
        s3_client = boto3.client(
            's3',
            aws_access_key_id='',
            aws_secret_access_key='',
            aws_session_token='',
        )

        client = s3.S3SyncClient(s3_client, 'testbucket', 'foo/bar')
        assert client.index_path() == 'foo/bar/.index'

    def test_put(self, s3_client):
        # given
        data = b'munchkin'

        # when
        input_object = SyncObject(io.BytesIO(data), len(data), 4000)
        s3_client.put('something/boardgame.rst', input_object)

        # then
        resp = s3_client.boto.get_object(
            Bucket=s3_client.bucket,
            Key=os.path.join(s3_client.prefix, 'something/boardgame.rst'),
        )
        assert resp['Body'].read() == data
        assert s3_client.get_remote_timestamp('something/boardgame.rst') == 4000

    def test_get(self, s3_client):
        # given
        data = b'#000000'
        frozen_time = datetime.datetime(2016, 10, 23, 10, 30, tzinfo=datetime.timezone.utc)
        with freezegun.freeze_time(frozen_time):
            s3_client.boto.put_object(
                Bucket=s3_client.bucket,
                Key=os.path.join(s3_client.prefix, 'black.color'),
                Body=data,
            )

        # when
        output_object = s3_client.get('black.color')

        # then
        assert output_object.fp.read() == data
        assert output_object.total_size == len(data)
        assert output_object.timestamp == to_timestamp(frozen_time)

    def test_get_non_existant(self, s3_client):
        assert s3_client.get('idontexist.md') is None

    def test_put_get(self, s3_client):
        # given
        data = b'woooooooooshhh'
        input_object = SyncObject(io.BytesIO(data), len(data), 4000)
        frozen_time = datetime.datetime(2016, 10, 23, 10, 30, tzinfo=datetime.timezone.utc)

        # when
        with freezegun.freeze_time(frozen_time):
            s3_client.put('something/woosh.gif', input_object)

        # then
        output_object = s3_client.get('something/woosh.gif')
        assert output_object.fp.read() == data
        assert output_object.timestamp == to_timestamp(frozen_time)

    def test_delete(self, s3_client):
        # given
        s3_client.boto.put_object(
            Bucket=s3_client.bucket,
            Key=os.path.join(s3_client.prefix, 'war.png'),
            Body='bang',
        )

        # when
        assert s3_client.delete('war.png') is True

        # then
        assert s3_client.get('war.png') is None
        with pytest.raises(ClientError) as exc:
            s3_client.boto.head_object(
                Bucket=s3_client.bucket,
                Key=os.path.join(s3_client.prefix, 'war.png')
            )
        assert exc.value.response['Error']['Code'] == '404'

    def test_delete_non_existant(self, s3_client):
        assert s3_client.delete('idontexist.png') is False

    def test_get_local_keys(self, s3_client):
        # given
        utils.set_s3_contents(s3_client, 'war.png')
        utils.set_s3_contents(s3_client, 'this/is/fine')
        utils.set_s3_contents(s3_client, '.index')
        utils.write_s3(
            s3_client.boto,
            s3_client.bucket,
            os.path.join('iamsomethingelse', 'ishouldnotshow.txt'),
        )

        # when
        actual_output = s3_client.get_local_keys()

        # then
        expected_output = ['war.png', 'this/is/fine']
        assert sorted(actual_output) == sorted(expected_output)

    def test_get_index_keys(self, s3_client):
        utils.set_s3_index(s3_client, {
            'cow': {
                'local_timestamp': 4000,
                'remote_timestamp': 3000,
            },
            'chicken': {
                'local_timestamp': 4000,
                'remote_timestamp': 3000,
            }
        })
        assert sorted(s3_client.get_index_keys()) == sorted(['cow', 'chicken'])

    def test_get_index_timestamps(self, s3_client):
        # given
        utils.set_s3_index(s3_client, {
            'hello': {
                'remote_timestamp': 1234,
                'local_timestamp': 1200,
            },
            'world': {
                'remote_timestamp': 5000,
            }
        })

        # then
        assert s3_client.get_remote_timestamp('hello') == 1234
        assert s3_client.get_index_local_timestamp('hello') == 1200

        assert s3_client.get_remote_timestamp('world') == 5000
        assert s3_client.get_index_local_timestamp('world') is None

    def test_get_all_index_timestamps(self, s3_client):
        # given
        utils.set_s3_index(s3_client, {
            'hello': {
                'local_timestamp': 1200,
            },
            'world': {
                'local_timestamp': 4000,
            }
        })

        # then
        expected_output = {
            'hello': 1200,
            'world': 4000,
        }
        actual_output = s3_client.get_all_index_local_timestamps()
        assert actual_output == expected_output

    def test_update_index(self, s3_client):
        # given
        utils.set_s3_index(s3_client, {
            'red': {
                'remote_timestamp': 1234,
                'local_timestamp': 1200,
            },
            'green': {
                'remote_timestamp': 5000,
            }
        })

        utils.set_s3_contents(s3_client, 'red', timestamp=5001)
        utils.set_s3_contents(s3_client, 'yellow', timestamp=1000)
        utils.set_s3_contents(s3_client, 'orange', timestamp=2000)

        # when
        s3_client.update_index()

        # then
        expected_index = {
            'red': {
                'remote_timestamp': 1234,
                'local_timestamp': 5001,
            },
            'green': {
                'remote_timestamp': 5000,
                'local_timestamp': None,
            },
            'yellow': {
                'remote_timestamp': None,
                'local_timestamp': 1000,
            },
            'orange': {
                'remote_timestamp': None,
                'local_timestamp': 2000,
            },
        }
        assert s3_client.index == expected_index

    def test_get_real_local_timestamp(self, s3_client):
        # given
        utils.set_s3_contents(s3_client, 'orange', timestamp=2000)

        # then
        assert s3_client.get_real_local_timestamp('orange') == 2000
        assert s3_client.get_real_local_timestamp('idontexist') is None

    def test_get_all_real_local_timestamps(self, s3_client):
        # given
        utils.set_s3_contents(s3_client, 'chocolate/oranges', timestamp=2600)
        utils.set_s3_contents(s3_client, 'gummy/bears', timestamp=2400)
        utils.set_s3_contents(s3_client, 'carrot_cake', timestamp=2000)
        utils.set_s3_contents(s3_client, '.index', timestamp=2043)
        s3_client.reload_index()

        # then
        expected_output = {
            'chocolate/oranges': 2600,
            'gummy/bears': 2400,
            'carrot_cake': 2000,
        }
        actual_output = s3_client.get_all_real_local_timestamps()
        assert actual_output == expected_output

    def test_set_index_timestamps(self, s3_client):
        # given
        utils.set_s3_index(s3_client, {
            'red': {
                'remote_timestamp': 1234,
                'local_timestamp': 1200,
            },
        })

        # when
        s3_client.set_index_local_timestamp('red', 3000)
        s3_client.set_remote_timestamp('red', 4000)
        s3_client.set_index_local_timestamp('green', 5000)
        s3_client.set_remote_timestamp('yellow', 6000)

        # then
        expected_index = {
            'red': {
                'local_timestamp': 3000,
                'remote_timestamp': 4000,
            },
            'green': {
                'local_timestamp': 5000,
            },
            'yellow': {
                'remote_timestamp': 6000,
            },
        }
        assert s3_client.index == expected_index

    def test_interrupted_put(self, s3_client):
        utils.set_s3_contents(s3_client, 'keychain', data='iamsomedatathatexists')

        sync_object = SyncObject(utils.InterruptedBytesIO(), 900000, 3000)

        with pytest.raises(ValueError):
            s3_client.put('keychain', sync_object)

        result = s3_client.get('keychain')
        assert result.fp.read() == b'iamsomedatathatexists'

    def test_ignore_files(self, s3_client):
        utils.set_s3_contents(
            s3_client, '.syncignore', timestamp=4000, data=(
                '*~\n'
                'trashdirectory\n'
                '*py[oc]\n'
                '.git\n'
            )
        )
        s3_client.reload_ignore_files()

        utils.set_s3_contents(s3_client, '.zshrc~')
        utils.set_s3_contents(s3_client, 'trashdirectory')
        utils.set_s3_contents(s3_client, 'foo/mobile.py', timestamp=1290)
        utils.set_s3_contents(s3_client, 'foo/mobile.pyc')
        utils.set_s3_contents(s3_client, 'foo/mobile.pyo')
        utils.set_s3_contents(s3_client, 'foo/.git/7f04e3b69963e5d51b479b30591ab1c2b6a3e7')
        utils.set_s3_contents(s3_client, '.zshrc', timestamp=9999)

        s3_client.update_index()

        expected_index = {
            '.syncignore': {
                'local_timestamp': 4000,
                'remote_timestamp': None,
            },
            '.zshrc': {
                'local_timestamp': 9999,
                'remote_timestamp': None,
            },
            'foo/mobile.py': {
                'local_timestamp': 1290,
                'remote_timestamp': None,
            }
        }
        assert s3_client.index == expected_index



# -*- coding: utf-8 -*-

import datetime
import getpass


def to_timestamp(dt):
    epoch = datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)
    return (dt - epoch) / datetime.timedelta(seconds=1)


def get_input(*args, secret=False, **kwargs):
    if secret:
        return getpass.getpass(*args, **kwargs)
    else:
        return input(*args, **kwargs)



# -*- coding: utf-8 -*-

import argparse
import io
import json
import logging
import os
import tempfile
from datetime import datetime

import mock
import pytest
import pytz

from s4 import cli
from s4.utils import to_timestamp
import utils


class FakeInputStream(object):
    def __init__(self, results):
        self.results = results
        self.index = 0

    def __call__(self, *args, **kwargs):
        output = self.results[self.index]
        self.index += 1
        return output


@pytest.yield_fixture
def config_file():
    fd, temp_path = tempfile.mkstemp()
    mocker = mock.patch('s4.cli.CONFIG_FILE_PATH', temp_path)
    mocker.start()
    yield temp_path
    mocker.stop()
    os.close(fd)
    os.remove(temp_path)


@pytest.fixture
def logger():
    stream = io.StringIO()
    result = logging.getLogger('test_targets')
    result.setLevel(logging.INFO)
    result.handlers = []
    result.addHandler(logging.StreamHandler(stream))
    return result


def get_stream_value(logger):
    return logger.handlers[0].stream.getvalue()


def get_timestamp(year, month, day, hour, minute):
    return to_timestamp(
        datetime(year, month, day, hour, minute, tzinfo=pytz.UTC)
    )


class TestMain(object):
    @mock.patch('argparse.ArgumentParser.print_help')
    def test_no_arguments_prints_help(self, print_help):
        cli.main([])
        assert print_help.call_count == 1

    def test_version_command(self, capsys):
        cli.main(['version'])
        out, err = capsys.readouterr()
        assert out == '{}\n'.format(cli.VERSION)

    @mock.patch('s4.cli.ls_command')
    def test_ls_command(self, ls_command):
        cli.main(['ls', 'foo'])
        assert ls_command.call_count == 1

    @mock.patch('s4.cli.sync_command')
    def test_sync_command(self, sync_command):
        cli.main(['sync', 'foo'])
        assert sync_command.call_count == 1

    @mock.patch('s4.cli.edit_command')
    def test_edit_command(self, edit_command):
        cli.main(['edit', 'foo'])
        assert edit_command.call_count == 1

    @mock.patch('s4.cli.targets_command')
    def test_targets_command(self, targets_command):
        cli.main(['targets'])
        assert targets_command.call_count == 1

    @mock.patch('s4.cli.rm_command')
    def test_rm_command(self, rm_command):
        cli.main(['rm', 'foo'])
        assert rm_command.call_count == 1

    @mock.patch('s4.cli.add_command')
    def test_add_command(self, add_command):
        cli.main(['add'])
        assert add_command.call_count == 1


class TestGetConfigFile(object):
    @mock.patch('s4.cli.CONFIG_FILE_PATH', '/i/dont/exist')
    def test_no_file(self):
        assert cli.get_config() == {}

    def test_correct_output(self, config_file):
        with open(config_file, 'w') as fp:
            json.dump({'local_folder': '/home/someone/something'}, fp)

        assert cli.get_config() == {'local_folder': '/home/someone/something'}

@mock.patch('s4.sync.SyncWorker')
class TestSyncCommand(object):
    def test_no_targets(self, SyncWorker, logger):
        args = argparse.Namespace(targets=None, conflicts=None)
        cli.sync_command(args, {'targets': {}}, logger)
        assert get_stream_value(logger) == ''
        assert SyncWorker.call_count == 0

    def test_wrong_target(self, SyncWorker, logger):
        args = argparse.Namespace(targets=['foo', 'bar'], conflicts=None)
        cli.sync_command(args, {'targets': {'baz': {}}}, logger)
        assert get_stream_value(logger) == (
            '"bar" is an unknown target. Choices are: [\'baz\']\n'
            '"foo" is an unknown target. Choices are: [\'baz\']\n'
        )
        assert SyncWorker.call_count == 0

    def test_keyboard_interrupt(self, SyncWorker, logger):
        args = argparse.Namespace(targets=None, conflicts=None)
        config = {
            'targets': {
                'foo': {
                    'local_folder': '/home/mike/docs',
                    's3_uri': 's3://foobar/docs',
                    'aws_access_key_id': '3223323',
                    'aws_secret_access_key': '23#@423#@',
                    'region_name': 'us-east-1',
                },
                'bar': {
                    'local_folder': '/home/mike/barmil',
                    's3_uri': 's3://foobar/barrel',
                    'aws_access_key_id': '3223',
                    'aws_secret_access_key': '23#eWEa@423#@',
                    'region_name': 'us-west-2',
                }
            }
        }
        SyncWorker.side_effect = KeyboardInterrupt

        cli.sync_command(args,  config, logger)
        assert SyncWorker.call_count == 1
        assert get_stream_value(logger) == (
            'Quitting due to Keyboard Interrupt...\n'
        )

    def test_all_targets(self, SyncWorker, logger):
        args = argparse.Namespace(targets=None, conflicts=None)
        config = {
            'targets': {
                'foo': {
                    'local_folder': '/home/mike/docs',
                    's3_uri': 's3://foobar/docs',
                    'aws_access_key_id': '3223323',
                    'aws_secret_access_key': '23#@423#@',
                    'region_name': 'us-east-1',
                },
                'bar': {
                    'local_folder': '/home/mike/barmil',
                    's3_uri': 's3://foobar/barrel',
                    'aws_access_key_id': '3223',
                    'aws_secret_access_key': '23#eWEa@423#@',
                    'region_name': 'us-west-2',
                }
            }
        }

        cli.sync_command(args,  config, logger)
        assert get_stream_value(logger) == (
            'Syncing bar [/home/mike/barmil/ <=> s3://foobar/barrel/]\n'
            'Syncing foo [/home/mike/docs/ <=> s3://foobar/docs/]\n'
        )
        assert SyncWorker.call_count == 2


@mock.patch('s4.utils.get_input')
class TestEditCommand(object):
    def test_no_targets(self, get_input, logger):
        cli.edit_command(None, {}, logger)
        expected_result = (
            'You have not added any targets yet\n'
            'Use the "add" command to do this\n'
        )
        assert get_stream_value(logger) == expected_result

    def test_missing_target(self, get_input, logger):
        args = argparse.Namespace(target='idontexist')
        config = {
            'targets': {'foo': {}}
        }
        cli.edit_command(args, config, logger)

        expected_result = (
            '"idontexist" is an unknown target\n'
            'Choices are: [\'foo\']\n'
        )
        assert get_stream_value(logger) == expected_result

    def test_no_changes(self, get_input, logger, config_file):
        fake_stream = FakeInputStream([
            '',
            '',
            '',
            '',
            '',
        ])
        get_input.side_effect = fake_stream

        args = argparse.Namespace(target='foo')
        config = {
            'targets': {'foo': {
                'local_folder': '/home/mark/documents',
                's3_uri': 's3://buckets/mybackup',
                'aws_access_key_id': '23123123123313',
                'aws_secret_access_key': 'edwdadwadwadwdd',
                'region_name': 'eu-west-1',
            }}
        }
        cli.edit_command(args, config, logger)

        with open(config_file, 'r') as fp:
            config = json.load(fp)

        expected_config = {
            'targets': {'foo': {
                'local_folder': '/home/mark/documents',
                's3_uri': 's3://buckets/mybackup',
                'aws_access_key_id': '23123123123313',
                'aws_secret_access_key': 'edwdadwadwadwdd',
                'region_name': 'eu-west-1',
            }}
        }
        assert expected_config == config


    def test_correct_output(self, get_input, logger, config_file):
        fake_stream = FakeInputStream([
            '/home/user/Documents',
            's3://buckets/mybackup222',
            '9999999999',
            'bbbbbbbbbbbbbbbbbbbbbbbb',
            'eu-west-2',
        ])
        get_input.side_effect = fake_stream

        args = argparse.Namespace(target='foo')
        config = {
            'targets': {'foo': {
                'local_folder': '/home/mark/documents',
                's3_uri': 's3://buckets/mybackup',
                'aws_access_key_id': '23123123123313',
                'aws_secret_access_key': 'edwdadwadwadwdd',
                'region_name': 'eu-west-1',
            }}
        }
        cli.edit_command(args, config, logger)

        with open(config_file, 'r') as fp:
            config = json.load(fp)

        expected_config = {
            'targets': {'foo': {
                'local_folder': '/home/user/Documents',
                's3_uri': 's3://buckets/mybackup222',
                'aws_access_key_id': '9999999999',
                'aws_secret_access_key': 'bbbbbbbbbbbbbbbbbbbbbbbb',
                'region_name': 'eu-west-2',
            }}
        }
        assert expected_config == config


@mock.patch('s4.utils.get_input')
class TestAddCommand(object):
    def test_correct_behaviour(self, get_input, logger, config_file):
        fake_stream = FakeInputStream([
            '/home/user/Documents',
            's3://mybucket/Documents',
            'aaaaaaaaaaaaaaaaaaaaaaaa',
            'bbbbbbbbbbbbbbbbbbbbbbbb',
            'eu-west-2',
            '',
        ])
        get_input.side_effect = fake_stream

        cli.add_command(None, {}, logger)

        with open(config_file, 'r') as fp:
            new_config = json.load(fp)

        expected_config = {
            'targets': {
                'Documents': {
                    'local_folder': '/home/user/Documents',
                    's3_uri': 's3://mybucket/Documents',
                    'aws_access_key_id': 'aaaaaaaaaaaaaaaaaaaaaaaa',
                    'aws_secret_access_key': 'bbbbbbbbbbbbbbbbbbbbbbbb',
                    'region_name': 'eu-west-2'
                }
            }
        }
        assert new_config == expected_config

    def test_custom_target_name(self, get_input, logger, config_file):
        fake_stream = FakeInputStream([
            '/home/user/Music',
            's3://mybucket/Musiccccc',
            '1234567890',
            'abcdefghij',
            'us-west-1',
            'Tunes',
        ])
        get_input.side_effect = fake_stream

        cli.add_command(None, {}, logger)

        with open(config_file, 'r') as fp:
            new_config = json.load(fp)

        expected_config = {
            'targets': {
                'Tunes': {
                    'local_folder': '/home/user/Music',
                    's3_uri': 's3://mybucket/Musiccccc',
                    'aws_access_key_id': '1234567890',
                    'aws_secret_access_key': 'abcdefghij',
                    'region_name': 'us-west-1'
                }
            }
        }
        assert new_config == expected_config


class TestLsCommand(object):
    def test_empty_config(self, logger):
        args = argparse.Namespace(target='idontexist')
        cli.ls_command(args, {}, logger)
        expected_result = (
            'You have not added any targets yet\n'
            'Use the "add" command to do this\n'
        )
        assert get_stream_value(logger) == expected_result

    def test_missing_target(self, logger):
        args = argparse.Namespace(target='idontexist')
        config = {
            'targets': {
                'foo': {},
                'bar': {},
            }
        }
        cli.ls_command(args, config, logger)
        expected_result = (
            '"idontexist" is an unknown target\n'
            'Choices are: [\'bar\', \'foo\']\n'
        )
        assert get_stream_value(logger) == expected_result

    def test_correct_output_empty(self, s3_client, local_client, logger):
        config = {
            'targets': {
                'foo': {
                    'local_folder': local_client.get_uri(),
                    's3_uri': s3_client.get_uri(),
                    'aws_access_key_id': '',
                    'aws_secret_access_key': '',
                    'region_name': 'eu-west-2',
                }
            }
        }

        args = argparse.Namespace(target='foo', sort_by='key', descending=False)
        cli.ls_command(args, config, logger)

        expected_result = (
            'key    local    s3\n'
            '-----  -------  ----\n'
        )
        assert get_stream_value(logger) == expected_result

    def test_correct_output_nonempty(self, s3_client, local_client, logger):
        config = {
            'targets': {
                'foo': {
                    'local_folder': local_client.get_uri(),
                    's3_uri': s3_client.get_uri(),
                    'aws_access_key_id': '',
                    'aws_secret_access_key': '',
                    'region_name': 'eu-west-2',
                }
            }
        }
        utils.set_s3_index(s3_client, {
            'milk': {
                'local_timestamp': get_timestamp(1989, 10, 23, 11, 30)
            },
            'honey': {
                'local_timestamp': get_timestamp(2016, 12, 12, 8, 30)
            },
            'ginger': {
                'local_timestamp': None,
            }
        })
        utils.set_local_index(local_client, {
            'milk': {
                'local_timestamp': get_timestamp(2016, 12, 12, 8, 30)
            },
            'honey': {
                'local_timestamp': get_timestamp(2016, 11, 10, 18, 40)
            },
            'lemon': {
                'local_timestamp': get_timestamp(2017, 2, 2, 8, 30)
            },
            'ginger': {
                'local_timestamp': None,
            }
        })

        args = argparse.Namespace(
            target='foo',
            sort_by='key',
            show_all=False,
            descending=False,
        )
        cli.ls_command(args, config, logger)

        expected_result = (
            'key    local                s3\n'
            '-----  -------------------  -------------------\n'
            'honey  2016-11-10 18:40:00  2016-12-12 08:30:00\n'
            'lemon  2017-02-02 08:30:00\n'
            'milk   2016-12-12 08:30:00  1989-10-23 11:30:00\n'

        )
        assert get_stream_value(logger) == expected_result

    def test_show_all(self, s3_client, local_client, logger):
        config = {
            'targets': {
                'foo': {
                    'local_folder': local_client.get_uri(),
                    's3_uri': s3_client.get_uri(),
                    'aws_access_key_id': '',
                    'aws_secret_access_key': '',
                    'region_name': 'eu-west-2',
                }
            }
        }
        utils.set_s3_index(s3_client, {
            'cheese': {
                'local_timestamp': get_timestamp(2017, 12, 12, 8, 30)
            },
            'crackers': {
                'local_timestamp': None,
            }
        })
        utils.set_local_index(local_client, {
            'cheese': {
                'local_timestamp': get_timestamp(2017, 2, 2, 8, 30)
            },
            'crackers': {
                'local_timestamp': None,
            }
        })

        args = argparse.Namespace(
            target='foo',
            sort_by='key',
            show_all=True,
            descending=False,
        )
        cli.ls_command(args, config, logger)

        expected_result = (
            'key       local                s3\n'
            '--------  -------------------  -------------------\n'
            'cheese    2017-02-02 08:30:00  2017-12-12 08:30:00\n'
            'crackers  <deleted>\n'

        )
        assert get_stream_value(logger) == expected_result



class TestTargetsCommand(object):

    def test_empty(self, logger):
        cli.targets_command(None, {}, logger)
        assert get_stream_value(logger) == ''

    def test_correct_output(self, logger):
        config = {
            'targets': {
                'Personal': {
                    's3_uri': 's3://mybackup/Personal',
                    'local_folder': '/home/user/Documents',
                },
                'Studies': {
                    's3_uri': 's3://something/something/Studies',
                    'local_folder': '/media/backup/Studies',
                },
            }
        }

        cli.targets_command(None, config, logger)

        expected_result = (
            'Personal: [/home/user/Documents <=> s3://mybackup/Personal]\n'
            'Studies: [/media/backup/Studies <=> s3://something/something/Studies]\n'
        )
        assert get_stream_value(logger) == expected_result


class TestRemoveCommand(object):
    def test_empty(self, logger, config_file):
        args = argparse.Namespace(target='foo')
        cli.rm_command(args, {}, logger)

        expected_output = (
            'You have not added any targets yet\n'
        )
        assert get_stream_value(logger) == expected_output

    def test_missing(self, logger, config_file):
        args = argparse.Namespace(target='foo')
        cli.rm_command(args, {
            'targets': {
                'bar': {}
            }
        }, logger)

        expected_output = (
            '"foo" is an unknown target\n'
            'Choices are: [\'bar\']\n'
        )
        assert get_stream_value(logger) == expected_output

    def test_remove_target(self, logger, config_file):
        args = argparse.Namespace(target='foo')
        cli.rm_command(args, {
            'targets': {
                'bar': {},
                'foo': {},
            }
        }, logger)

        assert get_stream_value(logger) == ''

        with open(config_file, 'rt') as fp:
            new_config = json.load(fp)

        expected_config = {'targets': {'bar': {}}}
        assert new_config == expected_config



# -*- coding: utf-8 -*-

import datetime

import pytest

from s4.clients import get_sync_state, SyncClient, SyncState, SyncObject


class TestSyncState(object):
    def test_repr(self):
        action = SyncState(SyncState.UPDATED, 1000, 53434)
        assert repr(action) == 'SyncState<UPDATED, local={}, remote={}>'.format(
            datetime.datetime.utcfromtimestamp(1000),
            datetime.datetime.utcfromtimestamp(53434)
        )

    def test_get_local_datetime_missing(self):
        action = SyncState(SyncState.UPDATED, None, None)
        assert action.get_local_datetime() is None

    def test_get_remote_datetime_missing(self):
        action = SyncState(SyncState.CREATED, None, None)
        assert action.get_remote_datetime() is None

    def test_equality_with_self(self):
        action = SyncState(SyncState.UPDATED, 1000, 53434)
        assert action == action

    def test_non_equality(self):
        assert SyncState(SyncState.UPDATED, 10, 30) != SyncState(SyncState.DELETED, 10, 30)
        assert SyncState(SyncState.DELETED, 20, 20) != SyncState(SyncState.DELETED, 40, 20)
        assert SyncState(SyncState.CONFLICT, 20, 20) != SyncState(SyncState.CONFLICT, 20, 40)

    def test_equality_with_wrong_type(self):
        action = SyncState(SyncState.DELETED, 1000, 3000)
        assert action != 10
        assert action != "hello"


class TestSyncClient(object):
    def test_get_client_name(self):
        client = SyncClient()
        with pytest.raises(NotImplementedError):
            client.get_client_name()

    def test_get_uri(self):
        client = SyncClient()
        with pytest.raises(NotImplementedError):
            client.get_uri()

    def test_put(self):
        client = SyncClient()
        with pytest.raises(NotImplementedError):
            client.put("something", None)

    def test_get(self):
        client = SyncClient()
        with pytest.raises(NotImplementedError):
            client.get("something")

    def test_delete(self):
        client = SyncClient()
        with pytest.raises(NotImplementedError):
            client.delete("something")

    def test_get_local_keys(self):
        client = SyncClient()
        with pytest.raises(NotImplementedError):
            client.get_local_keys()

    def test_get_real_local_timestamp(self):
        client = SyncClient()
        with pytest.raises(NotImplementedError):
            client.get_real_local_timestamp("something")

    def test_get_index_keys(self):
        client = SyncClient()
        with pytest.raises(NotImplementedError):
            client.get_index_keys()

    def test_get_index_local_timestamp(self):
        client = SyncClient()
        with pytest.raises(NotImplementedError):
            client.get_index_local_timestamp("something")

    def test_set_index_local_timestamp(self):
        client = SyncClient()
        with pytest.raises(NotImplementedError):
            client.set_index_local_timestamp("something", None)

    def test_get_remote_timestamp(self):
        client = SyncClient()
        with pytest.raises(NotImplementedError):
            client.get_remote_timestamp("something")

    def test_get_all_remote_timestamps(self):
        client = SyncClient()
        with pytest.raises(NotImplementedError):
            client.get_all_remote_timestamps()

    def test_get_all_index_local_timestamps(self):
        client = SyncClient()
        with pytest.raises(NotImplementedError):
            client.get_all_index_local_timestamps()

    def test_get_all_real_local_timestamps(self):
        client = SyncClient()
        with pytest.raises(NotImplementedError):
            client.get_all_real_local_timestamps()

    def test_flush_index(self):
        client = SyncClient()
        with pytest.raises(NotImplementedError):
            client.flush_index()


class TestSyncObject(object):
    def test_repr(self):
        dev_null = open('/dev/null', 'r')
        sync_object = SyncObject(
            fp=dev_null,
            total_size=4096,
            timestamp=312313,
        )
        assert sync_object.fp == dev_null
        assert sync_object.total_size == 4096
        assert sync_object.timestamp == 312313

        expected_repr = 'SyncObject<{}, 4096, 312313>'.format(dev_null)
        assert repr(sync_object) == expected_repr


class TestGetSyncState(object):
    def test_does_not_exist(self):
        assert get_sync_state(None, None, None) == SyncState(SyncState.DOESNOTEXIST, None, None)

    def test_no_changes(self):
        actual_state = get_sync_state(
            index_local=8130,
            real_local=8130,
            remote=8130,
        )
        expected_state = SyncState(SyncState.NOCHANGES, 8130, 8130)
        assert actual_state == expected_state

    def test_deleted(self):
        actual_state = get_sync_state(
            index_local=90000,
            real_local=None,
            remote=90000,
        )
        expected_state = SyncState(SyncState.DELETED, None, 90000)
        assert actual_state == expected_state

    def test_already_deleted(self):
        actual_state = get_sync_state(
            index_local=None,
            real_local=None,
            remote=77777,
        )
        expected_state = SyncState(SyncState.DELETED, None, 77777)
        assert actual_state == expected_state

    def test_updated(self):
        actual_state = get_sync_state(
            index_local=5000,
            real_local=6000,
            remote=5000,
        )
        expected_state = SyncState(SyncState.UPDATED, 6000, 5000)
        assert actual_state == expected_state

    def test_created(self):
        actual_state = get_sync_state(
            index_local=None,
            real_local=6000,
            remote=None,
        )
        expected_state = SyncState(SyncState.CREATED, 6000, None)
        assert actual_state == expected_state

    def test_conflict(self):
        actual_state = get_sync_state(
            index_local=6000,
            real_local=5000,
            remote=6000,
        )
        expected_state = SyncState(SyncState.CONFLICT, 6000, 6000)
        assert actual_state == expected_state

    def test_ignores_floating_precision(self):
        actual_state = get_sync_state(
            index_local=8000.80,
            real_local=8000.32,
            remote=6000,
        )
        expected_state = SyncState(SyncState.NOCHANGES, 8000, 6000)
        assert actual_state == expected_state



# -*- coding: utf-8 -*-
import collections
import fnmatch
import json
import logging
import os
import zlib

from botocore.exceptions import ClientError

import magic

from s4 import utils
from s4.clients import SyncClient, SyncObject


logger = logging.getLogger(__name__)


S3Uri = collections.namedtuple('S3Uri', ['bucket', 'key'])


def parse_s3_uri(uri):
    if not uri.startswith('s3://'):
        return None

    tokens = uri.replace('s3://', '').split('/')
    if len(tokens) < 2:
        return None

    bucket = tokens[0]
    key = '/'.join(tokens[1:])
    return S3Uri(bucket, key)


def is_ignored_key(key, ignore_files):
    # Check if any subdirectories match the ignore patterns
    key_parts = key.split('/')
    for part in key_parts:
        if any(fnmatch.fnmatch(part, pattern) for pattern in ignore_files):
            return True
    else:
        return False


class S3SyncClient(SyncClient):
    def __init__(self, boto, bucket, prefix):
        self.boto = boto
        self.bucket = bucket
        self.prefix = prefix
        self.index = self.load_index()
        self.reload_ignore_files()

    def get_client_name(self):
        return 's3'

    def __repr__(self):
        return 'S3SyncClient<{}, {}>'.format(self.bucket, self.prefix)

    def get_uri(self, key=''):
        return 's3://{}/{}'.format(self.bucket, os.path.join(self.prefix, key))

    def index_path(self):
        return os.path.join(self.prefix, '.index')

    def put(self, key, sync_object, callback=None):
        self.boto.upload_fileobj(
            Bucket=self.bucket,
            Key=os.path.join(self.prefix, key),
            Fileobj=sync_object.fp,
            Callback=callback,
        )
        self.set_remote_timestamp(key, sync_object.timestamp)

    def get(self, key):
        try:
            resp = self.boto.get_object(
                Bucket=self.bucket,
                Key=os.path.join(self.prefix, key),
            )
            return SyncObject(
                resp['Body'],
                resp['ContentLength'],
                utils.to_timestamp(resp['LastModified']),
            )
        except ClientError:
            return None

    def delete(self, key):
        resp = self.boto.delete_objects(
            Bucket=self.bucket,
            Delete={
                'Objects': [{'Key': os.path.join(self.prefix, key)}]
            }
        )
        return 'Deleted' in resp

    def load_index(self):
        try:
            resp = self.boto.get_object(
                Bucket=self.bucket,
                Key=self.index_path(),
            )
            body = resp['Body'].read()
            content_type = magic.from_buffer(body, mime=True)
            if content_type == 'text/plain':
                logger.debug('Detected plain text encoding for index')
                return json.loads(body.decode('utf-8'))
            elif content_type == 'application/zlib':
                logger.debug('Detected zlib encoding for index')
                body = zlib.decompress(body)
                return json.loads(body.decode('utf-8'))
            elif content_type == 'application/x-empty':
                return {}
            else:
                raise ValueError('Unknown content type for index', content_type)
        except (ClientError):
            return {}

    def reload_index(self):
        self.index = self.load_index()

    def flush_index(self, compressed=True):
        data = json.dumps(self.index).encode('utf-8')
        if compressed:
            logger.debug('Using zlib encoding for writing index')
            data = zlib.compress(data)
        else:
            logger.debug('Using plain text encoding for writing index')

        self.boto.put_object(
            Bucket=self.bucket,
            Key=self.index_path(),
            Body=data,
        )

    def get_local_keys(self):
        results = []
        paginator = self.boto.get_paginator('list_objects_v2')
        page_iterator = paginator.paginate(
            Bucket=self.bucket,
            Prefix=self.prefix,
        )
        for page in page_iterator:
            if 'Contents' not in page:
                return results

            for obj in page['Contents']:
                key = os.path.relpath(obj['Key'], self.prefix)
                if not is_ignored_key(key, self.ignore_files):
                    results.append(key)
                else:
                    logger.debug('Ignoring %s', key)

        return results

    def get_real_local_timestamp(self, key):
        try:
            response = self.boto.head_object(
                Bucket=self.bucket,
                Key=os.path.join(self.prefix, key),
            )
            return utils.to_timestamp(response['LastModified'])
        except ClientError:
            return None

    def get_index_keys(self):
        return self.index.keys()

    def get_index_local_timestamp(self, key):
        return self.index.get(key, {}).get('local_timestamp')

    def set_index_local_timestamp(self, key, timestamp):
        if key not in self.index:
            self.index[key] = {}
        self.index[key]['local_timestamp'] = timestamp

    def get_remote_timestamp(self, key):
        return self.index.get(key, {}).get('remote_timestamp')

    def set_remote_timestamp(self, key, timestamp):
        if key not in self.index:
            self.index[key] = {}
        self.index[key]['remote_timestamp'] = timestamp

    def get_all_real_local_timestamps(self):
        result = {}
        paginator = self.boto.get_paginator('list_objects_v2')
        page_iterator = paginator.paginate(
            Bucket=self.bucket,
            Prefix=self.prefix,
        )
        for page in page_iterator:
            for obj in page.get('Contents', []):
                key = os.path.relpath(obj['Key'], self.prefix)
                if not is_ignored_key(key, self.ignore_files):
                    result[key] = utils.to_timestamp(obj['LastModified'])

        return result

    def get_all_remote_timestamps(self):
        return {key: value.get('remote_timestamp') for key, value in self.index.items()}

    def get_all_index_local_timestamps(self):
        return {key: value.get('local_timestamp') for key, value in self.index.items()}

    def reload_ignore_files(self):
        self.ignore_files = ['.index']
        try:
            response = self.boto.get_object(
                Bucket=self.bucket,
                Key=os.path.join(self.prefix, '.syncignore')
            )
            data = response['Body'].read()
            data = data.decode('utf8')
            ignore_list = data.split('\n')
            self.ignore_files.extend(ignore_list)
        except ClientError:
            pass
